/*********************************************************************
*                SEGGER Microcontroller GmbH & Co. KG                *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2017 SEGGER Microcontroller GmbH & Co. KG         *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: 4.32                                             *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : RTOSInit_RL78G1A.c  for RENESAS RL78 CPU
Purpose : Initializes and handles the hardware for embOS as far
          as required by embOS.
          Feel free to modify this file acc. to your target system.
--------  END-OF-HEADER  ---------------------------------------------
*/

#include "RTOS.h"
#ifdef __ICCRL78__
  #include <ior5f10ele.h>
  #include <ior5f10ele_ext.h>
#elif defined (__GNUC__)
  #include "iodefine.h"
  #include "iodefine_ext.h"
#elif defined (__CCRL__)
  #include "iodefine.h"
#endif

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/

#ifndef   USE_WATCHDOG
  #define USE_WATCHDOG     (0)
#endif

#ifndef   OPT_BYTES_SIZE
  #define OPT_BYTES_SIZE   (4uL)
#endif

#ifndef   SECU_ID_SIZE
  #define SECU_ID_SIZE     (10uL)
#endif

#ifndef   OS_USE_INT_STACK
  #define OS_USE_INT_STACK (1)
#endif

/*********************************************************************
*
*       Clock frequency settings (configuration)
*/
#ifndef   OS_FSYS                   /* CPU Main clock frequency     */
  #define OS_FSYS 32000000uL        /* Assume 32MHz                 */
#endif

#ifndef   OS_PCLK_TIMER             /* Peripheral clock for timer   */
  #define OS_PCLK_TIMER (OS_FSYS)   /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_PCLK_UART              /* Peripheral clock for UART    */
  #define OS_PCLK_UART (OS_FSYS)    /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_TICK_FREQ
  #define OS_TICK_FREQ (1000u)
#endif

#define   OS_TIMER_RELOAD (OS_PCLK_TIMER/OS_TICK_FREQ)

/*********************************************************************
*
*       Configuration of communication to embOSView
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_UART
#endif

/*********************************************************************
*
*       UART settings for embOSView
*       If you do not want (or can not due to hardware limitations)
*       to dedicate a UART to embOSView, please define it to be -1
*       Currently the standard code enables UART1 per default
*       and supports UART1 only
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #ifndef   OS_UART
    #define OS_UART (1)
  #endif

  #ifndef   OS_BAUDRATE
    #define OS_BAUDRATE (38400uL)
  #endif
#endif

/****** End of configurable options *********************************/

/*********************************************************************
*
*       CPU configuration,
*       Option bytes and security code
*/
#ifdef __ICCRL78__
  #pragma diag_suppress=Ta026

  #if USE_WATCHDOG
  #pragma location = "OPTBYTE"
  __root __far const unsigned char OptionByte[OPT_BYTES_SIZE] = {
    //  0xFEu, 0xFFu, 0xE8u, 0x85u,  // This version requires correct ID bytes
    0xFEu, 0xFFu, 0xE8u, 0x84u,      // This version erases automatically when wrong ID bytes are sent
  };

  #else

  #pragma location = "OPTBYTE"
  __root __far const unsigned char OptionByte[OPT_BYTES_SIZE] = {
    //  0x6Eu, 0xFFu, 0xE8u, 0x85u,  // This version requires correct ID bytes
    0x6Eu, 0xFFu, 0xE8u, 0x84u,    // This version erases automatically when wrong ID bytes are sent
  };
  #endif // USE_WATCHDOG

  /* Security Byte Definition                                                  */
  #pragma location = "SECUID"
  __root __far const unsigned char SecuIDCode[SECU_ID_SIZE] = {
    0xffu, 0xffu, 0xffu, 0xffu, 0xffu, 0xffu, 0xffu, 0xffu, 0xffu, 0xffu
  };

  #define _NOP() __no_operation()

#elif defined (__GNUC__)
  //
  // Please ensure that sections ".option_bytes" and ".security_id" are defined in linker
  // and ensure that the option bytes are defined in vector_table.c
  //
  #define _NOP() __asm("nop")
#elif defined (__CCRL__)
  //
  // Please ensure the Security ID is defined in linker settings
  // and ensure the ".option_byte" section is defined in cstart.asm
  //
  #define _NOP() __nop()
#else
  #error "Undefined Compiler, Optionbytes have to be defined !!!"
#endif // Compiler specific option byte definition

/*********************************************************************
*
*       Check configuration
*
**********************************************************************
*/
#ifndef   DEBUG     /* Should normally be defined as project option */
  #define DEBUG  (0)
#endif

/*********************************************************************
*
*       OS interrupt stack
*/
#if OS_USE_INT_STACK
  #define OS_ENTER_INT_STACK() OS_EnterIntStack()
  #define OS_LEAVE_INT_STACK() OS_LeaveIntStack()
#else
  #define OS_ENTER_INT_STACK()
  #define OS_LEAVE_INT_STACK()
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
* Function description
*   Returns the current hardware timer count value
*
* Return value
*   Current timer count value
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return TCR00;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
* Function description
*   Returns if the hardware timer interrupt pending flag is set
*
* Return value
*   == 0; Interrupt pending flag not set
*   != 0: Interrupt pending flag set
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return TMIF00;
}


/*********************************************************************
*
*       OS_ISR_Tick()
*
*  Function description
*    This is the code that gets called when the processor receives a
*    INTTM00 exception. TM00 is used as OS timer tick.
*/
#ifdef __ICCRL78__
  #define OS_TICKVECT   INTTM00_vect
  #pragma vector=OS_TICKVECT
  static __interrupt __v1_call void OS_ISR_Tick(void) {
#elif defined (__CCRL__)
  #pragma interrupt OS_ISR_Tick (vect=INTTM00)
  static void OS_ISR_Tick (void) {
#elif defined (__GNUC__)
  void __attribute__ ((interrupt)) OS_ISR_Tick(void) {
#endif
  OS_EnterNestableInterrupt();
  OS_ENTER_INT_STACK();
  OS_TICK_Handle();
  OS_LEAVE_INT_STACK();
  OS_LeaveNestableInterrupt();
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/
/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*       Initialize the hardware (timer) required for embOS to run.
*       May be modified, if an other timer should be used
*/
void OS_InitHW(void) {
  //
  // Structure with information about timer frequency, tick frequency, etc.
  // for micro second precise system time.
  //
  OS_SYSTIMER_CONFIG SysTimerConfig = {OS_PCLK_TIMER, OS_TICK_FREQ, 0, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};
  OS_Config_SysTimer(&SysTimerConfig);
  //
  // Initially disable interrupts
  //
  OS_IncDI();
  //
  // Enable Timer array unit clock first
  // This is required before programming any other timer array sfr
  //
  TAU0EN  = 1u;                 // Enable Timer array 0 clock
  TT0    |= 0x0001;             // Stop timer channel 0
  TMMK00  = 1u;                 // Disable INTTM00 interrupt
  TMIF00  = 0u;                 // Clear INTTM00 interrupt flag
  TMPR100 = 1u;                 // Set INTTM00 to lowest priority
  TMPR000 = 1u;

  TMR00   =  0x0000u;           // Set interval timer mode for channel 0, no interrupt at start
  TDR00   = OS_TIMER_RELOAD - 1u;
  TO0    &= ~0x0001u;           // Clear timer output
  TOE0   &= ~0x0001u;           // Disable timer output signal
  //
  // Finally enable timer interrupt and start timer
  //
  TMIF00  = 0u;                 // Clear timer interrupt flag
  TMMK00 &= 0u;                 // Enable timer interrupt
  TS0    |= 0x0001u;            // Start timer
  //
  // Initialize the optional communication for embOSView
  //
#if (OS_VIEW_IFSELECT != OS_VIEW_DISABLED)
  OS_COM_Init();
#endif
  OS_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*       This is basically the "core" of the idle loop.
*       This core loop can be changed, but:
*       The idle loop does not have a stack of its own, therefore no
*       functionality should be implemented that relies on the stack
*       to be preserved. However, a simple program loop can be programmed
*       (like toggling an output or incrementing a counter)
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
                        // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       OS_GetTime_Cycles()
*
*  Function description
*       This routine is required for task-info via embOSView or high
*       resolution time measurement functions.
*       It returns the system time in timer clock cycles.
*/
OS_U32 OS_GetTime_Cycles(void) {
  OS_U32 Time;
  OS_U16 Cnt;

  Time = OS_GetTime32();
  Cnt  = OS_TIMER_RELOAD - TCR00;
  //
  // Check if timer interrupt pending ...
  //
  if (TMIF00 != 0) {
    //
    // We missed one interrupt, correct result
    //
    Cnt = OS_TIMER_RELOAD - TCR00;
    Time++;
  }
  return (OS_TIMER_RELOAD * Time) + Cnt;
}

/*********************************************************************
*
*       OS_ConvertCycles2us()
*
*  Function description
*       Convert Cycles into micro seconds.
*
*       If your clock frequency is not a multiple of 1 MHz,
*       you may have to modify this routine in order to get proper
*       diagnostics.
*
*       This routine is required for profiling or high resolution time
*       measurement only. It does not affect operation of the OS.
*/
OS_U32 OS_ConvertCycles2us(OS_U32 Cycles) {
  return (Cycles / (OS_TIMER_RELOAD / OS_TICK_FREQ));
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)

#if (OS_PCLK_UART == 32000000uL)
  //
  // Define UART clock prescaler value depending on baudrate setting
  //
  #if   (OS_BAUDRATE == 1200)
    #define UART_CLOCK_PRESCALE_VALUE 7
  #elif (OS_BAUDRATE == 2400)
    #define UART_CLOCK_PRESCALE_VALUE 6
  #elif (OS_BAUDRATE == 4800)
    #define UART_CLOCK_PRESCALE_VALUE 5
  #elif (OS_BAUDRATE == 9600)
    #define UART_CLOCK_PRESCALE_VALUE 4
  #elif (OS_BAUDRATE == 19200)
    #define UART_CLOCK_PRESCALE_VALUE 3
  #elif (OS_BAUDRATE == 38400)
    #define UART_CLOCK_PRESCALE_VALUE 2
  #else
    #error "Selected baudrate not supported."
  #endif // OS_BAUDRATE
#else
  #error "Baudrate calculation is based on 32MHz and has to be modified!"
#endif

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character via UART

*/
void OS_COM_Send1(OS_U8 c) {
  TXD1 = c;     // Send current character
}

/*********************************************************************
*
*       _OS_COM_ISR_Rx()
*
*  Function description
*    Interrupt handler for UART reception
*/
#ifdef __ICCRL78__
  #pragma vector = (INTSR1_vect)
  static __interrupt __v1_call void _OS_COM_ISR_Rx(void);
  static __interrupt __v1_call void _OS_COM_ISR_Rx(void) {
#elif defined (__GNUC__)
  void __attribute__ ((interrupt)) OS_COM_ISR_Rx(void) {
#elif defined (__CCRL__)
  #pragma interrupt OS_COM_ISR_Rx (vect=INTSR2)
  static void OS_COM_ISR_Rx (void) {
#endif
  OS_U8 RxData;

  OS_EnterNestableInterrupt();  /* We will enable interrupts   */
  OS_ENTER_INT_STACK();
  RxData = RXD1;
  OS_OnRx(RxData);
  OS_LEAVE_INT_STACK();
  OS_LeaveNestableInterrupt();
}

/*********************************************************************
*
*       _OS_COM_ISR_RxErr()
*
*  Function description
*    Interrupt handler for UART reception error
*/
#ifdef __ICCRL78__
  #pragma vector = (INTSRE1_vect)
  static __interrupt __v1_call void _OS_COM_ISR_RxErr(void);
  static __interrupt __v1_call void _OS_COM_ISR_RxErr(void) {
#elif defined (__GNUC__)
  void __attribute__ ((interrupt)) OS_COM_ISR_RxErr(void) {
#elif defined (__CCRL__)
  #pragma interrupt OS_COM_ISR_RxErr (vect=INTSRE2)
  static void OS_COM_ISR_RxErr(void) {
#endif
  volatile OS_U8 ErrData;

  OS_EnterNestableInterrupt();  /* We will enable interrupts   */
  OS_ENTER_INT_STACK();
  ErrData = RXD1;               /* Perform a dummy data read   */
  ErrData = (OS_U8)(SSR03 & 0x0007U);  /* Read error status    */
  SIR03   = (OS_U16)ErrData;           /* Reset error          */
  OS_LEAVE_INT_STACK();
  OS_LeaveNestableInterrupt();
}

/*********************************************************************
*
*       _OS_COM_ISR_Tx()
*
*  Function description
*    Interrupt handler for UART transmission
*/
#ifdef __ICCRL78__
  #pragma vector = (INTST1_vect)
  static __interrupt __v1_call void _OS_COM_ISR_Tx(void);
  static __interrupt __v1_call void _OS_COM_ISR_Tx(void) {
#elif defined (__GNUC__)
  void __attribute__ ((interrupt)) OS_COM_ISR_Tx(void) {
#elif defined (__CCRL__)
  #pragma interrupt OS_COM_ISR_Tx (vect=INTST2)
  static void OS_COM_ISR_Tx(void) {
#endif
  OS_EnterNestableInterrupt();  /* We will enable interrupts  */
  OS_ENTER_INT_STACK();
  OS_OnTx();
  OS_LEAVE_INT_STACK();
  OS_LeaveNestableInterrupt();
}

/*********************************************************************
*
*       OS_COM_Init()
*
*  Function description
*    Initializes Uart communication for embOSView
*/
void OS_COM_Init(void) {
  PER0 |= (1uL << 2);   // Enable clock of UART unit
  _NOP();               // Add a delay of 4 fclock cycles
  _NOP();
  _NOP();
  _NOP();
  //
  // Setup the UART clock prescaler to achieve selected baudrate
  //
  SPS0 = ((UART_CLOCK_PRESCALE_VALUE) | ((UART_CLOCK_PRESCALE_VALUE) << 4));
  //
  // Stop the UART and reset all interrupts
  //
  ST0    |= ((1uL << 2) | (1uL << 3));  // Stop UART1 receive and transmit
  STMK1   = 1u;  // Disable TXD1 interrupt
  STIF1   = 0u;  // Clear TxD1 interrupt flag
  SRMK1   = 1u;  // Disable RxD1 interrupt
  SRIF1   = 0u;  // Clear RxD1 interrupt flag
  SREMK1  = 1u;  // Disable RxD1Error interrupt
  SREIF1  = 0u;  // Clear RxD1Error interrupt flag
  //
  // Set interrupt priorities to low level
  //
  STPR11  = 1u;  // TxD1 interrupt priority
  STPR01  = 1u;
  SRPR11  = 1u;  // RxD1 interrupt priority
  SRPR01  = 1u;
  SREPR11 = 1u;  // RxD1 error interrupt priority
  SREPR01 = 1u;

  SMR02 = (1uL << 0)    // MD0: 0 => Transfer end interrupt, 1 => Buffer empty interrupt
        | (1uL << 1)    // MD2..MD1: 0 => CSI, 1 => UART, 2 => I2C
        | (1uL << 5)    // b5:  Always 1 !
        | (0uL << 6)    // SIS: 0 => falling edge is start bit
        | (0uL << 8)    // STS: 0 => Only software trigger (for UART)
        | (0uL << 14)   // CCS: 0 => internal clock used
        | (0ul << 15)   // CKS: Operation clock CKm0 set by SPSm register
        ;
  //
  // Setup channel 2, UART1 TxD
  //
  SCR02 = (3uL << 0)    // b1..b0: DLS, 3 => 8 bit data
        | (1uL << 2)    // b2:     Always 1 !
        | (0uL << 3)    // b3:     Always 0 !
        | (1uL << 4)    // b5..b4: SLC, 0 => No Stop, 1 => 1 Stop bit, 2 => 2 Stop bits
        | (0uL << 6)    // b6:     Always 0 !
        | (1uL << 7)    // b7:     DIR, 1 => LSB first
        | (0uL << 8)    // b9..b8  Parity, 0 => NONE
        | (0uL << 10)   // b10:    EOC, 0 => Mask error interrupt
        | (0uL << 11)   // b11:    Always 0 !
        | (0uL << 12)   // b12:    Clock phase, has to be 0 in UART mode !
        | (0uL << 13)   // b13:    DAP, has to be 0 in UART mode !
        | (0uL << 14)   // b14:    Rx enable, disabled
        | (1uL << 15)   // b15:    Tx enable
        ;
  SDR02 = (103uL << 9); // Fixed baudrate generator value for transmission

  //
  // Setup channel 3, UART1 RxD
  //
  NFEN0 |=(1uL << 2);   // Enable noise filter for RxD1
  SIR03 = (1uL << 0)    // Clear Overflow error state
        | (1uL << 1)    // Clear parity error flag
        | (1uL << 2)    // Clear Framing error flag
        ;
  SMR03 = (0uL << 0)    // MD0: 0 => Transfer end interrupt, 1 => Buffer empty interrupt
        | (1uL << 1)    // MD2..MD1: 0 => CSI, 1 => UART, 2 => I2C
        | (1uL << 5)    // b5:  Always 1 !
        | (0uL << 6)    // SIS: 0 => falling edge is start bit
        | (1uL << 8)    // STS: 1 => Valid edge of the RxD pin
        | (0uL << 14)   // CCS: 0 => internal clock used
        | (0ul << 15)   // CKS: Operation clock CKm0 set by SPSm register
        ;
  SCR03 = (3uL << 0)    // b1..b0: DLS, 3 => 8bit data
        | (1uL << 2)    // b2:     Always 1 !
        | (0uL << 3)    // b3:     Always 0 !
        | (1uL << 4)    // b5..b4: SLC, 0 => No Stop, 1 => 1 Stop bit, 2 => 2 Stop bits
        | (0uL << 6)    // b6:     Always 0 !
        | (1uL << 7)    // b7:     DIR, 1 => LSB first
        | (0uL << 8)    // b9..b8  Parity, 0 => NONE
        | (1uL << 10)   // b10:    EOC, 1 => Allow Rx error interrupt
        | (0uL << 11)   // b11:    Always 0 !
        | (0uL << 12)   // b12:    Clock phase, has to be 0 in UART mode !
        | (0uL << 13)   // b13:    DAP, has to be 0 in UART mode !
        | (1uL << 14)   // b14:    Rx enable
        | (0uL << 15)   // b15:    Tx enable, disabled
        ;
  SDR03 = (103uL << 9); // Fixed baudrate generator value for reception

  //
  // Setup port for UART communication
  //
  SO0  |=  (1uL << 2);  // Set UART TxD1 output to high level
  SOL0 &= ~(1uL << 2);  // Set UART normal (non inverted) output level
  SOE0 |=  (1uL << 2);  // Enable UART TxD1 output

  PMC0 &= ~(1uL << 3);  // Set RxD1 pin to UART input mode
  PM0  |=  (1uL << 3);
  PMC0 &= ~(1uL << 2);  // Set TxD1 pin to UART output mode
  P0   |=  (1uL << 2);  // Initially high
  PM0  &= ~(1uL << 2);
  //
  // Finally start the UART, enable interrupts
  //
  STIF1   = 0u;         // Clear TxD1 interrupt flag
  STMK1   = 0u;         // Enable TxD1 interrupt
  SRIF1   = 0u;         // Clear RxD1 interrupt flag
  SRMK1   = 0u;         // Enable RxD1 interrupt
  SREIF1  = 0u;         // Clear RxD1Error interrupt flag
  SREMK1  = 0u;         // Enable RxD1Error interrupt
  SS0    |= ((1uL << 3) | (1uL << 2));  // Enable UART1 Rx and Tx (channel 3 and 2)
}
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Dummy routine.
*/
void OS_COM_Send1(OS_U8 c) {
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
}

#else
  #error "Selected embOSView interface is currently not supported."
#endif

/*********************************************************************
*
*       HardwareSetup()
*/
#ifdef __cplusplus
extern "C" {
#endif
void HardwareSetup(void);
#ifdef __cplusplus
}
#endif
void HardwareSetup(void) {
  //
  // Speed up system clock
  //
  CMC     = 0x00u;    // Clock operation mode: All external clocks used as port
  MSTOP   = 1u;       // Stop external Oscillator X1
  //
  // Select internal high speed oscillator as clock
  //
  MCM0    = 0u;
  //
  // Select internal low speed clock as source for RTC, power RTC in stop mode
  //
  OSMC    = 0x10u;
  //
  // Activate main system clock as CPU-clock and peripheral clock
  //
  CSS     = 0u;
  HIOSTOP = 0u;       // Activate internal high-speed oscillator
  XTSTOP  = 1u;       // Stop XT1 oscillator
  //
  // Setup watchdog
  //
  WDTIMK  = 1u;       // Disable Watchdog interrupt
  WDTIIF  = 0u;       // Clear interrupt flag
#if USE_WATCHDOG
  WDTIPR1 = 0u;       // Set high priority interrupt priority
  WDTIPR0 = 0u;
  WDTIMK  = 0u;       // Enable Watchdog interrupt
#endif
  //
  // Some safety settings
  //
  CRC0CTL = 0x00U;    // Disable CRC calculation
  IAWCTL  = 0x80U;    // Enable invalid memory access detection
  //
  // Initial setup of timer array (channel 0 used as embOS tick)
  // Any modification here will require a modification in OS_InitHW()
  //
  TAU0EN  = 1u;       // Enable Timer array 0 clock
  TT0     = 0x0AFFu;  // Stop all timer channels
  TPS0    = 0x0000u;  // Clock frequency prescaler = 0 (Common for all channels)
  //
  // Perform further hardware setup here if required
  //
}

#if defined(__ICCRL78__)
/*********************************************************************
*
*       __low_level_init()
*       Perform hardware initialization which has to be done
*       before memory is initialized
*       Because this function is called, before variables are initialized,
*       no functions that rely on any variables can be performed
*/
#ifdef __cplusplus
extern "C" {
#endif
int __low_level_init(void);    // Avoid "no prototype" warning
#ifdef __cplusplus
  }
#endif
int __low_level_init(void) {
  HardwareSetup();
  return 1;           // Always return 1 to enable segment initialization
}
#endif  // __ICCRL78__

/****** End Of File *************************************************/
